/***************************************************************************
                          AtomicOperations.h  -  description
                             -------------------
    begin                : Feb 26, 2020
    copyright            : (C) 2020 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

// Implemented by: Tomas Oberhuber, Jakub Klinkovsky

#pragma once

#ifdef HAVE_CUDA
#include <cuda.h>
#endif
#ifdef HAVE_HIP
#include <hip/hip_runtime.h>
#endif

#include <TNL/Devices/DeviceCallable.h>
#include <TNL/Devices/Sequential.h>
#include <TNL/Devices/Host.h>
#include <TNL/Devices/Cuda.h>
#include <TNL/Devices/Hip.h>

namespace TNL {
namespace Algorithms {

template< typename Device >
struct AtomicOperations{};

template<>
struct AtomicOperations< Devices::Host >
{
   template< typename Value >
   static void add( Value& v, const Value& a )
   {
#pragma omp atomic update
      v += a;
   }
};

template<>
struct AtomicOperations< Devices::Cuda >
{
   template< typename Value >
   __device_callable__
   static void add( Value& v, const Value& a )
   {
#ifdef HAVE_CUDA
      atomicAdd( &v, a );
#endif // HAVE_CUDA
   }

#ifdef HAVE_CUDA
   __device__
   static void add( double& v, const double& a )
   {
#if __CUDA_ARCH__ < 600
      unsigned long long int* v_as_ull = ( unsigned long long int* ) &v;
      unsigned long long int old = *v_as_ull, assumed;

      do
      {
         assumed = old;
         old = atomicCAS( v_as_ull,
                          assumed,
                          __double_as_longlong( a + __longlong_as_double( assumed ) ) ) ;

      // Note: uses integer comparison to avoid hang in case of NaN (since NaN != NaN)
      }
      while( assumed != old );
#else // __CUDA_ARCH__ < 600
      atomicAdd( &v, a );
#endif //__CUDA_ARCH__ < 600
   }
#else // HAVE_CUDA
   static void add( double& v, const double& a ){}
#endif // HAVE_CUDA

   __device_callable__
   static void add( long int& v, const long int& a )
   {
#ifdef HAVE_CUDA
      TNL_ASSERT_TRUE( false, "Atomic add for long int is not supported on CUDA." );
#endif // HAVE_CUDA
   }

   __device_callable__
   static void add( short int& v, const short int& a )
   {
#ifdef HAVE_CUDA
      TNL_ASSERT_TRUE( false, "Atomic add for short int is not supported on CUDA." );
#endif // HAVE_CUDA
   }
};

template<>
struct AtomicOperations< Devices::Hip >
{
   template< typename Value >
   __device_callable__
   static void add( Value& v, const Value& a )
   {
#ifdef HAVE_HIP
      atomicAdd( &v, a );
#endif // HAVE_CUDA
   }

#ifdef HAVE_HIP
   __device__
   static void add( double& v, const double& a )
   {
#ifdef UNDEF // TODO: Use this when HIP supports atomic double.
      atomicAdd( &v, a );
#else
      // HIP currently does not supoport atomic operations for double:
      // "HIP enables atomic operations on 32-bit integers. Additionally, it supports
      // an atomic float add. AMD hardware, however, implements the float add using
      // a CAS loop, so this function may not perform efficiently."
      // See.: https://rocmdocs.amd.com/en/latest/Programming_Guides/HIP-GUIDE.html
      unsigned long long int* v_as_ull = ( unsigned long long int* ) &v;
      unsigned long long int old = *v_as_ull, assumed;

      do
      {
         assumed = old;
         old = atomicCAS( v_as_ull,
                          assumed,
                          __double_as_longlong( a + __longlong_as_double( assumed ) ) ) ;

      // Note: uses integer comparison to avoid hang in case of NaN (since NaN != NaN)
      }
      while( assumed != old );
#endif
   }
#else // HAVE_HIP
   static void add( double& v, const double& a ){}
#endif // HAVE_HIP

   __device_callable__
   static void add( long int& v, const long int& a )
   {
#ifdef HAVE_HIP
      TNL_ASSERT_TRUE( false, "Atomic add for long int is not supported on HIP." );
#endif // HAVE_HIP
   }

   __device_callable__
   static void add( short int& v, const short int& a )
   {
#ifdef HAVE_HIP
      TNL_ASSERT_TRUE( false, "Atomic add for short int is not supported on HIP." );
#endif // HAVE_HIP
   }
};

} //namespace Algorithms
} //namespace TNL
