/***************************************************************************
                          HipReductionKernel.h  -  description
                             -------------------
    begin                : Jun 18, 2020
    copyright            : (C) 2020 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <utility>  // std::pair

#ifdef HAVE_HIP
#include "hip/hip_runtime.h"
#endif

#include <TNL/Assert.h>
#include <TNL/Math.h>
#include <TNL/Hip/DeviceInfo.h>
#include <TNL/Hip/SharedMemory.h>
#include <TNL/Algorithms/HipReductionBuffer.h>
#include <TNL/Algorithms/MultiDeviceMemoryOperations.h>
#include <TNL/Exceptions/HipSupportMissing.h>

namespace TNL {
namespace Algorithms {

/****
 * In CUDA, the performance of this kernel is very sensitive to register usage.
 * Compile with --ptxas-options=-v and configure these constants for given
 * architecture so that there are no local memory spills.
 * TODO: Check this for HIP and ROCm.
 */
static constexpr int HipReduction_maxThreadsPerBlock = 256;  // must be a power of 2
static constexpr int HipReduction_registersPerThread = 32;   // empirically determined optimal value

#ifdef HAVE_HIP
// In CUDA, the following constant depends on GPU architecture.
static constexpr int HipReduction_minBlocksPerMultiprocessor = 8;


template< int blockSize,
          typename Result,
          typename DataFetcher,
          typename Reduction,
          typename Index >
__global__ void
__launch_bounds__( HipReduction_maxThreadsPerBlock, HipReduction_minBlocksPerMultiprocessor )
HipReductionKernel( const Result zero,
                     DataFetcher dataFetcher,
                     const Reduction reduction,
                     const Index begin,
                     const Index end,
                     Result* output )
{
   Result* sdata = Hip::getSharedMemory< Result >();

   // Get the thread id (tid), global thread id (gid) and gridSize.
   const Index tid = threadIdx.x;
         Index gid = begin + blockIdx.x * blockDim. x + threadIdx.x;
   const Index gridSize = blockDim.x * gridDim.x;

   sdata[ tid ] = zero;

   // Start with the sequential reduction and push the result into the shared memory.
   while( gid + 4 * gridSize < end ) {
      sdata[ tid ] = reduction( sdata[ tid ], dataFetcher( gid ) );
      sdata[ tid ] = reduction( sdata[ tid ], dataFetcher( gid + gridSize ) );
      sdata[ tid ] = reduction( sdata[ tid ], dataFetcher( gid + 2 * gridSize ) );
      sdata[ tid ] = reduction( sdata[ tid ], dataFetcher( gid + 3 * gridSize ) );
      gid += 4 * gridSize;
   }
   while( gid + 2 * gridSize < end ) {
      sdata[ tid ] = reduction( sdata[ tid ], dataFetcher( gid ) );
      sdata[ tid ] = reduction( sdata[ tid ], dataFetcher( gid + gridSize ) );
      gid += 2 * gridSize;
   }
   while( gid < end ) {
      sdata[ tid ] = reduction( sdata[ tid ], dataFetcher( gid ) );
      gid += gridSize;
   }
   __syncthreads();

   // Perform the parallel reduction.
   if( blockSize >= 1024 ) {
      if( tid < 512 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 512 ] );
      __syncthreads();
   }
   if( blockSize >= 512 ) {
      if( tid < 256 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 256 ] );
      __syncthreads();
   }
   if( blockSize >= 256 ) {
      if( tid < 128 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 128 ] );
      __syncthreads();
   }
   if( blockSize >= 128 ) {
      if( tid <  64 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 64 ] );
      __syncthreads();
   }

   // This runs in one warp so we use __syncwarp() instead of __syncthreads().
   if( tid < 32 ) {
      if( blockSize >= 64 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 32 ] );
      __syncthreads(); // TODO: Replace with __syncwarp when its supported by HIP
      //__syncwarp();
      // Note that here we do not have to check if tid < 16 etc, because we have
      // 2 * blockSize.x elements of shared memory per block, so we do not
      // access out of bounds. The results for the upper half will be undefined,
      // but unused anyway.
      if( blockSize >= 32 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 16 ] );
      __syncthreads(); // TODO: Replace with __syncwarp when its supported by HIP
      //__syncwarp();
      if( blockSize >= 16 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 8 ] );
      __syncthreads(); // TODO: Replace with __syncwarp when its supported by HIP
      //__syncwarp();
      if( blockSize >=  8 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 4 ] );
      __syncthreads(); // TODO: Replace with __syncwarp when its supported by HIP
      //__syncwarp();
      if( blockSize >=  4 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 2 ] );
      __syncthreads(); // TODO: Replace with __syncwarp when its supported by HIP
      //__syncwarp();
      if( blockSize >=  2 )
         sdata[ tid ] = reduction( sdata[ tid ], sdata[ tid + 1 ] );
   }

   // Store the result back in the global memory.
   if( tid == 0 )
      output[ blockIdx.x ] = sdata[ 0 ];
}

template< int blockSize,
          typename Result,
          typename DataFetcher,
          typename Reduction,
          typename Index >
__global__ void
__launch_bounds__( HipReduction_maxThreadsPerBlock, HipReduction_minBlocksPerMultiprocessor )
HipReductionWithArgumentKernel( const Result zero,
                                 DataFetcher dataFetcher,
                                 const Reduction reduction,
                                 const Index begin,
                                 const Index end,
                                 Result* output,
                                 Index* idxOutput,
                                 const Index* idxInput = nullptr )
{
   Result* sdata = Hip::getSharedMemory< Result >();
   Index* sidx = reinterpret_cast< Index* >( &sdata[ blockDim.x ] );

   // Get the thread id (tid), global thread id (gid) and gridSize.
   const Index tid = threadIdx.x;
         Index gid = begin + blockIdx.x * blockDim. x + threadIdx.x;
   const Index gridSize = blockDim.x * gridDim.x;

   // Start with the sequential reduction and push the result into the shared memory.
   if( idxInput ) {
      if( gid < end ) {
         sdata[ tid ] = dataFetcher( gid );
         sidx[ tid ] = idxInput[ gid ];
         gid += gridSize;
      } else {
         sdata[ tid ] = zero;
      }
      while( gid + 4 * gridSize < end ) {
         reduction( sdata[ tid ], dataFetcher( gid ), sidx[ tid ], idxInput[ gid ] );
         reduction( sdata[ tid ], dataFetcher( gid + gridSize ), sidx[ tid ], idxInput[ gid + gridSize ] );
         reduction( sdata[ tid ], dataFetcher( gid + 2 * gridSize ), sidx[ tid ], idxInput[ gid + 2 * gridSize ] );
         reduction( sdata[ tid ], dataFetcher( gid + 3 * gridSize ), sidx[ tid ], idxInput[ gid + 3 * gridSize ] );
         gid += 4 * gridSize;
      }
      while( gid + 2 * gridSize < end ) {
         reduction( sdata[ tid ], dataFetcher( gid ), sidx[ tid ], idxInput[ gid ] );
         reduction( sdata[ tid ], dataFetcher( gid + gridSize ), sidx[ tid ], idxInput[ gid + gridSize ] );
         gid += 2 * gridSize;
      }
      while( gid < end ) {
         reduction( sdata[ tid ], dataFetcher( gid ), sidx[ tid ], idxInput[ gid ] );
         gid += gridSize;
      }
   }
   else {
      if( gid < end ) {
         sdata[ tid ] = dataFetcher( gid );
         sidx[ tid ] = gid;
         gid += gridSize;
      } else {
         sdata[ tid ] = zero;
      }
      while( gid + 4 * gridSize < end ) {
         reduction( sdata[ tid ], dataFetcher( gid ), sidx[ tid ], gid );
         reduction( sdata[ tid ], dataFetcher( gid + gridSize ), sidx[ tid ], gid + gridSize );
         reduction( sdata[ tid ], dataFetcher( gid + 2 * gridSize ), sidx[ tid ], gid + 2 * gridSize );
         reduction( sdata[ tid ], dataFetcher( gid + 3 * gridSize ), sidx[ tid ], gid + 3 * gridSize );
         gid += 4 * gridSize;
      }
      while( gid + 2 * gridSize < end ) {
         reduction( sdata[ tid ], dataFetcher( gid ), sidx[ tid ], gid );
         reduction( sdata[ tid ], dataFetcher( gid + gridSize ), sidx[ tid ], gid + gridSize );
         gid += 2 * gridSize;
      }
      while( gid < end ) {
         reduction( sdata[ tid ], dataFetcher( gid ), sidx[ tid ], gid );
         gid += gridSize;
      }
   }
   __syncthreads();

   // Perform the parallel reduction.
   if( blockSize >= 1024 ) {
      if( tid < 512 )
         reduction( sdata[ tid ], sdata[ tid + 512 ], sidx[ tid ], sidx[ tid + 512 ] );
      __syncthreads();
   }
   if( blockSize >= 512 ) {
      if( tid < 256 )
         reduction( sdata[ tid ], sdata[ tid + 256 ], sidx[ tid ], sidx[ tid + 256 ] );
      __syncthreads();
   }
   if( blockSize >= 256 ) {
      if( tid < 128 )
         reduction( sdata[ tid ], sdata[ tid + 128 ], sidx[ tid ], sidx[ tid + 128 ] );
      __syncthreads();
   }
   if( blockSize >= 128 ) {
      if( tid <  64 )
         reduction( sdata[ tid ], sdata[ tid + 64 ], sidx[ tid ], sidx[ tid + 64 ] );
      __syncthreads();
   }

   // This runs in one warp so we use __syncwarp() instead of __syncthreads().
   if( tid < 32 ) {
      if( blockSize >= 64 )
         reduction( sdata[ tid ], sdata[ tid + 32 ], sidx[ tid ], sidx[ tid + 32 ] );
      __syncthreads(); // TODO: Replace with __syncwarp() when HIP supports it.
      //__syncwarp();
      // Note that here we do not have to check if tid < 16 etc, because we have
      // 2 * blockSize.x elements of shared memory per block, so we do not
      // access out of bounds. The results for the upper half will be undefined,
      // but unused anyway.
      if( blockSize >= 32 )
         reduction( sdata[ tid ], sdata[ tid + 16 ], sidx[ tid ], sidx[ tid + 16 ] );
      __syncthreads(); // TODO: Replace with __syncwarp() when HIP supports it.
      //__syncwarp();
      if( blockSize >= 16 )
         reduction( sdata[ tid ], sdata[ tid + 8 ], sidx[ tid ], sidx[ tid + 8 ] );
      __syncthreads(); // TODO: Replace with __syncwarp() when HIP supports it.
      //__syncwarp();
      if( blockSize >=  8 )
         reduction( sdata[ tid ], sdata[ tid + 4 ], sidx[ tid ], sidx[ tid + 4 ] );
      __syncthreads(); // TODO: Replace with __syncwarp() when HIP supports it.
      //__syncwarp();
      if( blockSize >=  4 )
         reduction( sdata[ tid ], sdata[ tid + 2 ], sidx[ tid ], sidx[ tid + 2 ] );
      __syncthreads(); // TODO: Replace with __syncwarp() when HIP supports it.
      //__syncwarp();
      if( blockSize >=  2 )
         reduction( sdata[ tid ], sdata[ tid + 1 ], sidx[ tid ], sidx[ tid + 1 ] );
   }

   // Store the result back in the global memory.
   if( tid == 0 ) {
      output[ blockIdx.x ] = sdata[ 0 ];
      idxOutput[ blockIdx.x ] = sidx[ 0 ];
   }
}
#endif


template< typename Index,
          typename Result >
struct HipReductionKernelLauncher
{
   // The number of blocks should be a multiple of the number of multiprocessors
   // to ensure optimum balancing of the load. This is very important, because
   // we run the kernel with a fixed number of blocks, so the amount of work per
   // block increases with enlarging the problem, so even small imbalance can
   // cost us dearly.
   // Therefore,  desGridSize = blocksPerMultiprocessor * numberOfMultiprocessors
   // where blocksPerMultiprocessor is determined according to the number of
   // available registers on the multiprocessor.
   // On Tesla K40c, desGridSize = 8 * 15 = 120.
   //
   // Update:
   // It seems to be better to map only one HIP block per one multiprocessor or maybe
   // just slightly more. Therefore we omit blocksdPerMultiprocessor in the following.
   HipReductionKernelLauncher( const Index begin, const Index end )
   : activeDevice( Hip::DeviceInfo::getActiveDevice() ),
     blocksdPerMultiprocessor( Hip::DeviceInfo::getRegistersPerMultiprocessor( activeDevice )
                               / ( HipReduction_maxThreadsPerBlock * HipReduction_registersPerThread ) ),
     //desGridSize( blocksdPerMultiprocessor * Hip::DeviceInfo::getHipMultiprocessors( activeDevice ) ),
     desGridSize( Hip::DeviceInfo::getHipMultiprocessors( activeDevice ) ),
     begin( begin ), end( end )
   {
   }

   template< typename DataFetcher,
             typename Reduction >
   int start( const Reduction& reduction,
              DataFetcher& dataFetcher,
              const Result& zero,
              Result*& output )
   {
      // create reference to the reduction buffer singleton and set size
      const std::size_t buf_size = 2 * desGridSize * sizeof( Result );
      HipReductionBuffer& hipReductionBuffer = HipReductionBuffer::getInstance();
      hipReductionBuffer.setSize( buf_size );
      output = hipReductionBuffer.template getData< Result >();

      this->reducedSize = this->launch( begin, end, reduction, dataFetcher, zero, output );
      return this->reducedSize;
   }

   template< typename DataFetcher,
             typename Reduction >
   int startWithArgument( const Reduction& reduction,
                          DataFetcher& dataFetcher,
                          const Result& zero,
                          Result*& output,
                          Index*& idxOutput )
   {
      // create reference to the reduction buffer singleton and set size
      const std::size_t buf_size = 2 * desGridSize * ( sizeof( Result ) + sizeof( Index ) );
      HipReductionBuffer& hipReductionBuffer = HipReductionBuffer::getInstance();
      hipReductionBuffer.setSize( buf_size );
      output = hipReductionBuffer.template getData< Result >();
      idxOutput = reinterpret_cast< Index* >( &output[ 2 * desGridSize ] );

      this->reducedSize = this->launchWithArgument( begin, end, reduction, dataFetcher, zero, output, idxOutput, nullptr );
      return this->reducedSize;
   }

   template< typename Reduction >
   Result
   finish( const Reduction& reduction,
           const Result& zero )
   {
      // Input is the first half of the buffer, output is the second half
      HipReductionBuffer& hipReductionBuffer = HipReductionBuffer::getInstance();
      Result* input = hipReductionBuffer.template getData< Result >();
      Result* output = &input[ desGridSize ];

      while( this->reducedSize > 1 )
      {
         // this lambda has to be defined inside the loop, because the captured variable changes
         auto copyFetch = [input] __device_callable__ ( Index i ) { return input[ i ]; };
         this->reducedSize = this->launch( 0, this->reducedSize, reduction, copyFetch, zero, output );
         std::swap( input, output );
      }

      // swap again to revert the swap from the last iteration
      // AND to solve the case when this->reducedSize was 1 since the beginning
      std::swap( input, output );

      // Copy result on CPU
      Result result;
      MultiDeviceMemoryOperations< void, Devices::Hip >::copy( &result, output, 1 );
      return result;
   }

   template< typename Reduction >
   std::pair< Result, Index >
   finishWithArgument( const Reduction& reduction,
                       const Result& zero )
   {
      // Input is the first half of the buffer, output is the second half
      HipReductionBuffer& hipReductionBuffer = HipReductionBuffer::getInstance();
      Result* input = hipReductionBuffer.template getData< Result >();
      Result* output = &input[ desGridSize ];
      Index* idxInput = reinterpret_cast< Index* >( &output[ desGridSize ] );
      Index* idxOutput = &idxInput[ desGridSize ];

      while( this->reducedSize > 1 )
      {
         // this lambda has to be defined inside the loop, because the captured variable changes
         auto copyFetch = [input] __device_callable__ ( Index i ) { return input[ i ]; };
         this->reducedSize = this->launchWithArgument( ( Index ) 0, this->reducedSize, reduction, copyFetch, zero, output, idxOutput, idxInput );
         std::swap( input, output );
         std::swap( idxInput, idxOutput );
      }

      // swap again to revert the swap from the last iteration
      // AND to solve the case when this->reducedSize was 1 since the beginning
      std::swap( input, output );
      std::swap( idxInput, idxOutput );

      ////
      // Copy result on CPU
      std::pair< Result, Index > result;
      MultiDeviceMemoryOperations< void, Devices::Hip >::copy( &result.first, output, 1 );
      MultiDeviceMemoryOperations< void, Devices::Hip >::copy( &result.second, idxOutput, 1 );
      return result;
   }


   protected:
      template< typename DataFetcher,
                typename Reduction >
      int launch( const Index begin,
                  const Index end,
                  const Reduction& reduction,
                  DataFetcher& dataFetcher,
                  const Result& zero,
                  Result* output )
      {
#ifdef HAVE_HIP
         const Index size = end - begin;
         dim3 blockSize, gridSize;
         blockSize.x = HipReduction_maxThreadsPerBlock;
         gridSize.x = TNL::min( Hip::getNumberOfBlocks( size, blockSize.x ), desGridSize );

         // when there is only one warp per blockSize.x, we need to allocate two warps
         // worth of shared memory so that we don't index shared memory out of bounds
         const Index shmem = (blockSize.x <= 32)
                  ? 2 * blockSize.x * sizeof( Result )
                  : blockSize.x * sizeof( Result );

         // This is "general", but this method always sets blockSize.x to a specific value,
         // so runtime switch is not necessary - it only prolongs the compilation time.
/*
         // Depending on the blockSize we generate appropriate template instance.
         switch( blockSize.x )
         {
            case 512:
               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel< 512 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
            case 256:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel< 256), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel< 256 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
            case 128:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel< 128), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel< 128 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
            case  64:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel<  64), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel<  64 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
            case  32:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel<  32), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel<  32 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
            case  16:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel<  16), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel<  16 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
           case   8:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel<   8), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel<   8 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
            case   4:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel<   4), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel<   4 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
            case   2:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel<   2), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel<   2 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output);
               break;
            case   1:
               TNL_ASSERT( false, std::cerr << "blockSize should not be 1." << std::endl );
            default:
               TNL_ASSERT( false, std::cerr << "Block size is " << blockSize. x << " which is none of 1, 2, 4, 8, 16, 32, 64, 128, 256 or 512." );
         }
         hipStreamSynchronize(0);
         TNL_CHECK_CUDA_DEVICE;
*/

         // Check just to future-proof the code setting blockSize.x
         if( blockSize.x == HipReduction_maxThreadsPerBlock ) {
            //hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionKernel< HipReduction_maxThreadsPerBlock, Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

            hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionKernel< HipReduction_maxThreadsPerBlock >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, begin, end, output);
            hipStreamSynchronize(0);
            TNL_CHECK_CUDA_DEVICE;
         }
         else {
            TNL_ASSERT( false, std::cerr << "Block size was expected to be " << HipReduction_maxThreadsPerBlock << ", but " << blockSize.x << " was specified." << std::endl; );
         }

         // Return the size of the output array on the CUDA device
         return gridSize.x;
#else
         throw Exceptions::HipSupportMissing();
#endif
      }

      template< typename DataFetcher,
                typename Reduction >
      int launchWithArgument( const Index begin,
                              const Index end,
                              const Reduction& reduction,
                              DataFetcher& dataFetcher,
                              const Result& zero,
                              Result* output,
                              Index* idxOutput,
                              const Index* idxInput )
      {
#ifdef HAVE_HIP
         dim3 blockSize, gridSize;
         const Index size = end - begin;
         blockSize.x = HipReduction_maxThreadsPerBlock;
         gridSize.x = TNL::min( Hip::getNumberOfBlocks( size, blockSize.x ), desGridSize );

         // when there is only one warp per blockSize.x, we need to allocate two warps
         // worth of shared memory so that we don't index shared memory out of bounds
         const Index shmem = (blockSize.x <= 32)
                  ? 2 * blockSize.x * ( sizeof( Result ) + sizeof( Index ) )
                  : blockSize.x * ( sizeof( Result ) + sizeof( Index ) );

         // This is "general", but this method always sets blockSize.x to a specific value,
         // so runtime switch is not necessary - it only prolongs the compilation time.
/*
         // Depending on the blockSize we generate appropriate template instance.
         switch( blockSize.x )
         {
            case 512:
               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel< 512 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
            case 256:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel< 256), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel< 256 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
            case 128:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel< 128), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel< 128 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
            case  64:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel<  64), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel<  64 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
            case  32:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel<  32), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel<  32 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
            case  16:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel<  16), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel<  16 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
           case   8:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel<   8), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel<   8 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
            case   4:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel<   4), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel<   4 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
            case   2:
               hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel<   2), Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

               hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel<   2 >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, size, output, idxOutput, idxInput );
               break;
            case   1:
               TNL_ASSERT( false, std::cerr << "blockSize should not be 1." << std::endl );
            default:
               TNL_ASSERT( false, std::cerr << "Block size is " << blockSize. x << " which is none of 1, 2, 4, 8, 16, 32, 64, 128, 256 or 512." );
         }
         hipStreamSynchronize(0);
         TNL_CHECK_CUDA_DEVICE;
*/

         // Check just to future-proof the code setting blockSize.x
         if( blockSize.x == HipReduction_maxThreadsPerBlock ) {
            //hipFuncSetCacheConfig(reinterpret_cast<const void*>(HipReductionWithArgumentKernel< HipReduction_maxThreadsPerBlock, Result, DataFetcher, Reduction, Index >, hipFuncCachePreferShared);

            hipLaunchKernelGGL(HIP_KERNEL_NAME(HipReductionWithArgumentKernel< HipReduction_maxThreadsPerBlock >), dim3(gridSize), dim3(blockSize), shmem , 0,  zero, dataFetcher, reduction, begin, end, output, idxOutput, idxInput );
            hipStreamSynchronize(0);
            TNL_CHECK_CUDA_DEVICE;
         }
         else {
            TNL_ASSERT( false, std::cerr << "Block size was expected to be " << HipReduction_maxThreadsPerBlock << ", but " << blockSize.x << " was specified." << std::endl; );
         }

         // return the size of the output array on the CUDA device
         return gridSize.x;
#else
         throw Exceptions::HipSupportMissing();
#endif
      }


      const int activeDevice;
      const int blocksdPerMultiprocessor;
      const int desGridSize;
      const Index begin, end;
      Index reducedSize;
};

} // namespace Algorithms
} // namespace TNL
