/***************************************************************************
                          Hip.h  -  description
                             -------------------
    begin                : Jun 18, 2020
    copyright            : (C) 2020 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

// Implemented by: Jakub Klinkovsky

#pragma once

#include <TNL/Exceptions/HipBadAlloc.h>
#include <TNL/Exceptions/HipSupportMissing.h>
#include <TNL/Hip/CheckDevice.h>

namespace TNL {
namespace Allocators {

/**
 * \brief Allocator for the HIP device memory space.
 *
 * The allocation is done using the `hipMalloc` function and the deallocation
 * is done using the `hipFree` function.
 */
template< class T >
struct Hip
{
   using value_type = T;
   using size_type = std::size_t;
   using difference_type = std::ptrdiff_t;

   Hip() = default;
   Hip( const Hip& ) = default;
   Hip( Hip&& ) = default;

   Hip& operator=( const Hip& ) = default;
   Hip& operator=( Hip&& ) = default;

   template< class U >
   Hip( const Hip< U >& )
   {}

   template< class U >
   Hip( Hip< U >&& )
   {}

   template< class U >
   Hip& operator=( const Hip< U >& )
   {
      return *this;
   }

   template< class U >
   Hip& operator=( Hip< U >&& )
   {
      return *this;
   }

   value_type* allocate( size_type n )
   {
#ifdef HAVE_HIP
      TNL_CHECK_HIP_DEVICE;
      value_type* result = nullptr;
      if( hipMalloc( (void**) &result, n * sizeof(value_type) ) != hipSuccess )
         throw Exceptions::HipBadAlloc();
      TNL_CHECK_HIP_DEVICE;
      return result;
#else
      throw Exceptions::HipSupportMissing();
#endif
   }

   void deallocate(value_type* ptr, size_type)
   {
#ifdef HAVE_HIP
      TNL_CHECK_HIP_DEVICE;
      hipFree( ptr );
      TNL_CHECK_HIP_DEVICE;
#else
      throw Exceptions::HipSupportMissing();
#endif
   }
};

template<class T1, class T2>
bool operator==(const Hip<T1>&, const Hip<T2>&)
{
   return true;
}

template<class T1, class T2>
bool operator!=(const Hip<T1>& lhs, const Hip<T2>& rhs)
{
   return !(lhs == rhs);
}

} // namespace Allocators
} // namespace TNL
