/***************************************************************************
                          HipHost.h  -  description
                             -------------------
    begin                : Apr 8, 2019
    copyright            : (C) 2019 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

// Implemented by: Jakub Klinkovsky

#pragma once

#include <TNL/Exceptions/HipBadAlloc.h>
#include <TNL/Exceptions/HipSupportMissing.h>
#include <TNL/Hip/CheckDevice.h>

namespace TNL {
namespace Allocators {

/**
 * \brief Allocator for page-locked memory on the host.
 *
 * The allocation is done using the `hipHostMalloc` function and the
 * deallocation is done using the `hipHostFree` function.
 */
template< class T >
struct HipHost
{
   using value_type = T;
   using size_type = std::size_t;
   using difference_type = std::ptrdiff_t;

   HipHost() = default;
   HipHost( const HipHost& ) = default;
   HipHost( HipHost&& ) = default;

   HipHost& operator=( const HipHost& ) = default;
   HipHost& operator=( HipHost&& ) = default;

   template< class U >
   HipHost( const HipHost< U >& )
   {}

   template< class U >
   HipHost( HipHost< U >&& )
   {}

   template< class U >
   HipHost& operator=( const HipHost< U >& )
   {
      return *this;
   }

   template< class U >
   HipHost& operator=( HipHost< U >&& )
   {
      return *this;
   }

   value_type* allocate( size_type n )
   {
#ifdef HAVE_HIP
      TNL_CHECK_HIP_DEVICE;
      value_type* result = nullptr;
      // hipHostMallocPortable - The memory returned by this call will be considered as pinned memory by all
      //                       HIP contexts, not just the one that performed the allocation.
      // hipHostMallocMapped - Maps the allocation into the HIP address space.
      if( hipHostMalloc( (void**) &result, n * sizeof(value_type), hipHostMallocPortable | hipHostMallocMapped ) != hipSuccess )
         throw Exceptions::HipBadAlloc();
      TNL_CHECK_HIP_DEVICE;
      return result;
#else
      throw Exceptions::HipSupportMissing();
#endif
   }

   void deallocate(value_type* ptr, size_type)
   {
#ifdef HAVE_HIP
      TNL_CHECK_HIP_DEVICE;
      hipHostFree( ptr );
      TNL_CHECK_HIP_DEVICE;
#else
      throw Exceptions::HipSupportMissing();
#endif
   }
};

template<class T1, class T2>
bool operator==(const HipHost<T1>&, const HipHost<T2>&)
{
   return true;
}

template<class T1, class T2>
bool operator!=(const HipHost<T1>& lhs, const HipHost<T2>& rhs)
{
   return !(lhs == rhs);
}

} // namespace Allocators
} // namespace TNL
