/***************************************************************************
                          HipManaged.h  -  description
                             -------------------
    begin                : Jun 18, 2020
    copyright            : (C) 2020 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

// Implemented by: Jakub Klinkovsky

#pragma once

#include <TNL/Exceptions/HipBadAlloc.h>
#include <TNL/Exceptions/HipSupportMissing.h>
#include <TNL/Hip/CheckDevice.h>

namespace TNL {
namespace Allocators {

/**
 * \brief Allocator for the HIP Unified Memory system.
 *
 * The memory allocated by this allocator will be automatically managed by the
 * HIP Unified Memory system. The allocation is done using the
 * `hipMallocManaged` function and the deallocation is done using the
 * `hipFree` function.
 * 
 */
template< class T >
struct HipManaged
{
   using value_type = T;
   using size_type = std::size_t;
   using difference_type = std::ptrdiff_t;

   HipManaged() = default;
   HipManaged( const HipManaged& ) = default;
   HipManaged( HipManaged&& ) = default;

   HipManaged& operator=( const HipManaged& ) = default;
   HipManaged& operator=( HipManaged&& ) = default;

   template< class U >
   HipManaged( const HipManaged< U >& )
   {}

   template< class U >
   HipManaged( HipManaged< U >&& )
   {}

   template< class U >
   HipManaged& operator=( const HipManaged< U >& )
   {
      return *this;
   }

   template< class U >
   HipManaged& operator=( HipManaged< U >&& )
   {
      return *this;
   }

   value_type* allocate( size_type n )
   {
#ifdef HAVE_HIP
      TNL_CHECK_HIP_DEVICE;
      value_type* result = nullptr;
      if( hipMallocManaged( &result, n * sizeof(value_type) ) != hipSuccess )
         throw Exceptions::HipBadAlloc();
      TNL_CHECK_HIP_DEVICE;
      return result;
#else
      throw Exceptions::HipSupportMissing();
#endif
   }

   void deallocate(value_type* ptr, size_type)
   {
#ifdef HAVE_HIP
      TNL_CHECK_HIP_DEVICE;
      hipFree( ptr );
      TNL_CHECK_HIP_DEVICE;
#else
      throw Exceptions::HipSupportMissing();
#endif
   }
};

template<class T1, class T2>
bool operator==(const HipManaged<T1>&, const HipManaged<T2>&)
{
   return true;
}

template<class T1, class T2>
bool operator!=(const HipManaged<T1>& lhs, const HipManaged<T2>& rhs)
{
   return !(lhs == rhs);
}

} // namespace Allocators
} // namespace TNL
