/***************************************************************************
                          DeviceInfo.hpp  -  description
                             -------------------
    begin                : Jun 21, 2015
    copyright            : (C) 2007 by Tomas Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <unordered_map>

#ifdef HAVE_HIP
#include <hip/hip_runtime.h>
#endif

#include <TNL/Hip/DeviceInfo.h>
#include <TNL/Exceptions/HipSupportMissing.h>

namespace TNL {
namespace Hip {

inline int
DeviceInfo::
getNumberOfDevices()
{
#ifdef HAVE_HIP
   int devices;
   hipGetDeviceCount( &devices );
   return devices;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getActiveDevice()
{
#ifdef HAVE_HIP
   int device;
   hipGetDevice( &device );
   return device;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline String
DeviceInfo::
getDeviceName( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return String( properties.name );
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getArchitectureMajor( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.major;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getArchitectureMinor( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.minor;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getClockRate( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.clockRate;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline std::size_t
DeviceInfo::
getGlobalMemory( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.totalGlobalMem;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline std::size_t
DeviceInfo::
getFreeGlobalMemory()
{
#ifdef HAVE_HIP
   std::size_t free = 0;
   std::size_t total = 0;
   hipMemGetInfo( &free, &total );
   return free;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getMemoryClockRate( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.memoryClockRate;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline bool
DeviceInfo::
getECCEnabled( int deviceNum )
{
#ifdef HAVE_HIP
   hipDeviceProp_t properties;
   hipGetDeviceProperties( &properties, deviceNum );
   return properties.ECCEnabled;
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getHipMultiprocessors( int deviceNum )
{
#ifdef HAVE_HIP
   // results are cached because they are used for configuration of some kernels
   static std::unordered_map< int, int > results;
   if( results.count( deviceNum ) == 0 ) {
      hipDeviceProp_t properties;
      hipGetDeviceProperties( &properties, deviceNum );
      results.emplace( deviceNum, properties.multiProcessorCount );
      return properties.multiProcessorCount;
   }
   return results[ deviceNum ];
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getHipCoresPerMultiprocessors( int deviceNum )
{
#ifdef HAVE_HIP
   return -1;
   // TODO: The following is very unlikely to work
   /*int major = DeviceInfo::getArchitectureMajor( deviceNum );
   int minor = DeviceInfo::getArchitectureMinor( deviceNum );
   switch( major )
   {
      case 1:   // Tesla generation, G80, G8x, G9x classes
         return 8;
      case 2:   // Fermi generation
         switch( minor )
         {
            case 0:  // GF100 class
               return 32;
            case 1:  // GF10x class
               return 48;
         }
      case 3: // Kepler generation -- GK10x, GK11x classes
         return 192;
      case 5: // Maxwell generation -- GM10x, GM20x classes
         return 128;
      case 6: // Pascal generation
         switch( minor )
         {
            case 0:  // GP100 class
               return 64;
            case 1:  // GP10x classes
            case 2:
               return 128;
         }
      default:
         return -1;
   }*/
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getHipCores( int deviceNum )
{
#ifdef HAVE_HIP
   return DeviceInfo::getHipMultiprocessors( deviceNum ) *
          DeviceInfo::getHipCoresPerMultiprocessors( deviceNum );
#else
   throw Exceptions::HipSupportMissing();
#endif
}

inline int
DeviceInfo::
getRegistersPerMultiprocessor( int deviceNum )
{
   // TODO: regsPerMultiprocessor is not part of hipDeviceProp_t yet.
   throw std::runtime_error( "HIP cannot detect number of registers per multiprocessor." );
#ifdef HAVE_HIP
   // results are cached because they are used for configuration of some kernels
   static std::unordered_map< int, int > results;
   if( results.count( deviceNum ) == 0 ) {
      hipDeviceProp_t properties;
      hipGetDeviceProperties( &properties, deviceNum );
      //results.emplace( deviceNum, properties.regsPerMultiprocessor );
      //return properties.regsPerMultiprocessor;
   }
   return results[ deviceNum ];
#else
   throw Exceptions::HipSupportMissing();
#endif
}

} // namespace Hip
} // namespace TNL
